@@ -2,6 +2,9 @@
 # published under the MPL-2.0 license)
 # https://github.com/abravalheri/ini2toml/blob/49897590a9254646434b7341225932e54f9626a3/LICENSE.txt
 
+# ruff: noqa: C408
+# Unnecessary `dict` call (rewrite as a literal)
+
 import argparse
 import io
 import json
@@ -13,6 +16,7 @@ from textwrap import dedent, wrap
 from typing import (
     Callable,
     Dict,
+    Generator,
     Iterator,
     List,
     NamedTuple,
@@ -23,32 +27,21 @@ from typing import (
 )
 
 from . import __version__
+from . import _tomllib as tomllib
 from .api import Validator
 from .errors import ValidationError
 from .plugins import PluginWrapper
 from .plugins import list_from_entry_points as list_plugins_from_entry_points
+from .remote import RemotePlugin, load_store
 
 _logger = logging.getLogger(__package__)
 T = TypeVar("T", bound=NamedTuple)
 
-
-try:  # pragma: no cover
-    if sys.version_info[:2] >= (3, 11):
-        from tomllib import TOMLDecodeError, loads
-    else:
-        from tomli import TOMLDecodeError, loads
-except ImportError:  # pragma: no cover
-    try:
-        from toml import TomlDecodeError as TOMLDecodeError  # type: ignore
-        from toml import loads  # type: ignore
-    except ImportError as ex:
-        raise ImportError("Please install `tomli` (TOML parser)") from ex
-
-_REGULAR_EXCEPTIONS = (ValidationError, TOMLDecodeError)
+_REGULAR_EXCEPTIONS = (ValidationError, tomllib.TOMLDecodeError)
 
 
 @contextmanager
-def critical_logging():
+def critical_logging() -> Generator[None, None, None]:
     """Make sure the logging level is set even before parsing the CLI args"""
     try:
         yield
@@ -108,12 +101,25 @@ META: Dict[str, dict] = {
         action="store_true",
         help="Print the JSON equivalent to the given TOML",
     ),
+    "tool": dict(
+        flags=("-t", "--tool"),
+        action="append",
+        dest="tool",
+        help="External tools file/url(s) to load, of the form name=URL#path",
+    ),
+    "store": dict(
+        flags=("--store",),
+        help="Load a pyproject.json file and read all the $ref's into tools "
+        "(see https://json.schemastore.org/pyproject.json)",
+    ),
 }
 
 
 class CliParams(NamedTuple):
     input_file: List[io.TextIOBase]
     plugins: List[PluginWrapper]
+    tool: List[str]
+    store: str
     loglevel: int = logging.WARNING
     dump_json: bool = False
 
@@ -121,7 +127,7 @@ class CliParams(NamedTuple):
 def __meta__(plugins: Sequence[PluginWrapper]) -> Dict[str, dict]:
     """'Hyper parameters' to instruct :mod:`argparse` how to create the CLI"""
     meta = {k: v.copy() for k, v in META.items()}
-    meta["enable"]["choices"] = set([p.tool for p in plugins])
+    meta["enable"]["choices"] = {p.tool for p in plugins}
     meta["input_file"]["default"] = [_STDIN]  # lazily defined to facilitate testing
     return meta
 
@@ -155,8 +161,10 @@ def parse_args(
     params = vars(parser.parse_args(args))
     enabled = params.pop("enable", ())
     disabled = params.pop("disable", ())
+    params["tool"] = params["tool"] or []
+    params["store"] = params["store"] or ""
     params["plugins"] = select_plugins(plugins, enabled, disabled)
-    return params_class(**params)
+    return params_class(**params)  # type: ignore[call-overload, no-any-return]
 
 
 def select_plugins(
@@ -172,7 +180,7 @@ def select_plugins(
     return available
 
 
-def setup_logging(loglevel: int):
+def setup_logging(loglevel: int) -> None:
     """Setup basic logging
 
     Args:
@@ -183,24 +191,24 @@ def setup_logging(loglevel: int):
 
 
 @contextmanager
-def exceptions2exit():
+def exceptions2exit() -> Generator[None, None, None]:
     try:
         yield
     except _ExceptionGroup as group:
         for prefix, ex in group:
             print(prefix)
             _logger.error(str(ex) + "\n")
-        raise SystemExit(1)
+        raise SystemExit(1) from None
     except _REGULAR_EXCEPTIONS as ex:
         _logger.error(str(ex))
-        raise SystemExit(1)
+        raise SystemExit(1) from None
     except Exception as ex:  # pragma: no cover
         _logger.error(f"{ex.__class__.__name__}: {ex}\n")
         _logger.debug("Please check the following information:", exc_info=True)
-        raise SystemExit(1)
+        raise SystemExit(1) from None
 
 
-def run(args: Sequence[str] = ()):
+def run(args: Sequence[str] = ()) -> int:
     """Wrapper allowing :obj:`Translator` to be called in a CLI fashion.
 
     Instead of returning the value from :func:`Translator.translate`, it prints the
@@ -214,7 +222,10 @@ def run(args: Sequence[str] = ()):
     plugins: List[PluginWrapper] = list_plugins_from_entry_points()
     params: CliParams = parse_args(args, plugins)
     setup_logging(params.loglevel)
-    validator = Validator(plugins=params.plugins)
+    tool_plugins = [RemotePlugin.from_str(t) for t in params.tool]
+    if params.store:
+        tool_plugins.extend(load_store(params.store))
+    validator = Validator(params.plugins, extra_plugins=tool_plugins)
 
     exceptions = _ExceptionGroup()
     for file in params.input_file:
@@ -228,11 +239,11 @@ def run(args: Sequence[str] = ()):
     return 0
 
 
-def _run_on_file(validator: Validator, params: CliParams, file: io.TextIOBase):
+def _run_on_file(validator: Validator, params: CliParams, file: io.TextIOBase) -> None:
     if file in (sys.stdin, _STDIN):
         print("Expecting input via `stdin`...", file=sys.stderr, flush=True)
 
-    toml_equivalent = loads(file.read())
+    toml_equivalent = tomllib.loads(file.read())
     validator(toml_equivalent)
     if params.dump_json:
         print(json.dumps(toml_equivalent, indent=2))
@@ -248,7 +259,7 @@ class Formatter(argparse.RawTextHelpFormatter):
     # order to create our own formatter, we are left no choice other then overwrite a
     # "private" method considered to be an implementation detail.
 
-    def _split_lines(self, text, width):
+    def _split_lines(self, text: str, width: int) -> List[str]:
         return list(chain.from_iterable(wrap(x, width) for x in text.splitlines()))
 
 
@@ -277,17 +288,17 @@ def _format_file(file: io.TextIOBase) -> str:
 class _ExceptionGroup(Exception):
     _members: List[Tuple[str, Exception]]
 
-    def __init__(self):
+    def __init__(self) -> None:
         self._members = []
         super().__init__()
 
-    def add(self, prefix: str, ex: Exception):
+    def add(self, prefix: str, ex: Exception) -> None:
         self._members.append((prefix, ex))
 
     def __iter__(self) -> Iterator[Tuple[str, Exception]]:
         return iter(self._members)
 
-    def raise_if_any(self):
+    def raise_if_any(self) -> None:
         number = len(self._members)
         if number == 1:
             print(self._members[0][0])
